from typing import List
from dramatist.drama.ScenePart import ScenePart


class StageDirection(ScenePart):
    _text: str = ''

    def __init__(self, text: str):
        self.tokens: List[str] = list()
        self.length = 0
        self.text = text

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, v: str):
        self._text = v
        self.tokens = v.split(' ')
        self.length = len(self.tokens)

    def __len__(self):
        return self.length



from dataclasses import dataclass, field
from typing import List, Tuple, Optional
import math

from dramatist.core.DramatistParseException import DramatistParseException
from dramatist.drama.Markup import Markup
from dramatist.drama.ScenePart import ScenePart
from dramatist.drama.Speech import Speech
from dramatist.drama.StageDirection import StageDirection
from dramatist.TextBlock import TextBlock


@dataclass
class Scene:
    title: str = ''
    scene_parts: List[ScenePart] = field(default_factory=list)
    start: int = 0
    end: int = 0
    start_line: int = 0
    end_line: int = 0
    markup: Markup = None

    def new_speech(self):
        self.scene_parts.append(Speech())

    def set_speaker(self, speaker: str):
        latest_block = self.scene_parts[-1]
        if isinstance(latest_block, Speech):
            latest_block.speaker = speaker
        else:
            raise DramatistParseException('Invalid latest block')

    def add_stage_direction(self, stage_direction: str):
        self.scene_parts.append(StageDirection(stage_direction))

    def add_speech(self, text: str, speech_type: int, markups: List[Markup]):
        latest_block = self.scene_parts[-1]
        if isinstance(latest_block, Speech):
            latest_block.add_speech_part(text, speech_type, markups)
        else:
            raise DramatistParseException('Invalid latest block')

    def get_text(self) -> str:
        text, _ = self.get_text_with_markup()
        return text

    def get_text_with_markup(self) -> Tuple[str, List[Markup]]:
        result = f'{self.title}'
        markups = []

        if self.markup:
            markups.append(self.markup)

        for scene_part in self.scene_parts:
            if isinstance(scene_part, StageDirection):
                result += f'\n{scene_part.text}'
            elif isinstance(scene_part, Speech):
                result += f'{scene_part.get_speaker()}{scene_part.get_text()}'
                markups.extend(scene_part.get_markups())

        return result, markups

    def get_text_in_blocks(self, max_length: int) -> List[TextBlock]:
        # Only returns spoken text and stage instructions in between.

        result: List[TextBlock] = []
        current: Optional[TextBlock] = None
        current_length: int = 0
        for scene_block in self.scene_parts:

            if isinstance(scene_block, StageDirection):
                continue
                # stage = scene_block
                #
                # if current:
                #     current.text += f'\n{stage.text}'
                # else:
                #     current = TextBlock(0, 0, stage.text)
                #
                # current_length += len(stage)
            elif isinstance(scene_block, Speech):
                speech = scene_block
                if current_length + len(speech) < max_length:
                    if current:
                        current.text += f'\n{speech.speaker}{speech.get_text()}'
                        current.end_line = speech.end_line
                    else:
                        current = TextBlock(speech.start_line, speech.end_line, f'{speech.speaker}{speech.get_text()}')

                    current_length += len(speech)
                else:
                    if current_length > 0:
                        result.append(current)
                        current = None
                        current_length = 0

                    if len(speech) < max_length:
                        current = TextBlock(speech.start_line, speech.end_line, f'{speech.speaker}{speech.get_text()}')
                        current_length = len(speech)
                    else:
                        factor = (len(speech) // max_length) + 1
                        sub_length = math.ceil(len(speech) / factor)
                        parts = [speech.get_tokens()[i:i + sub_length] for i in range(0, len(speech), sub_length)]

                        # Line numbers are only estimations for now
                        line_count = speech.end_line - speech.start_line + 1
                        lines_per_part = line_count // factor
                        line_start = speech.start_line

                        for pos, p in enumerate(parts):
                            text = ' '.join(p)

                            if pos == 0:
                                result.append(TextBlock(line_start, line_start + lines_per_part, f'{speech.speaker} {text}'))
                            else:
                                result.append(TextBlock(line_start, line_start + lines_per_part, text))

                            line_start += lines_per_part

        if current_length > 0:
            result.append(current)

        return result

    def get_text_by_speaker(self, max_length: int) -> List[TextBlock]:
        result: List[TextBlock] = []

        for scene_block in self.scene_parts:
            if isinstance(scene_block, Speech):
                speech = scene_block

                if len(speech) < max_length:
                    result.append(TextBlock(speech.start_line, speech.end_line, f'{speech.speaker}{speech.get_text()}'))
                else:
                    factor = (len(speech) // max_length) + 1
                    sub_length = math.ceil(len(speech) / factor)
                    parts = [speech.get_tokens()[i:i + sub_length] for i in range(0, len(speech), sub_length)]

                    # Line numbers are only estimations for now
                    line_count = speech.end_line - speech.start_line + 1
                    lines_per_part = line_count // factor
                    line_start = speech.start_line

                    for pos, p in enumerate(parts):
                        text = ' '.join(p)

                        if pos == 0:
                            result.append(TextBlock(line_start, line_start + lines_per_part, f'{speech.speaker}{text}'))
                        else:
                            result.append(TextBlock(line_start, line_start + lines_per_part, text))

                        line_start += lines_per_part

        return result

    def get_range(self) -> Tuple[int, int]:
        return self.start, self.end

    def get_line_range(self) -> Tuple[int, int]:
        return self.start_line, self.end_line

    def __len__(self):
        return len(self.scene_parts)



from dataclasses import dataclass, field
from typing import List

from dramatist.drama.Markup import Markup
from dramatist.drama.ScenePart import ScenePart
from dramatist.drama.SpeechPart import SpeechPart


@dataclass
class Speech(ScenePart):
    start_line: int = -1
    end_line: int = -1
    speaker: str = ''
    speaker_markup: Markup = None
    speech_parts: List[SpeechPart] = field(default_factory=list)

    def get_text(self) -> str:
        result = ''

        for pos, sb in enumerate(self.speech_parts):
            if sb.type == SpeechPart.TYPE_SPEECH_P_B:
                result += '\n'
            elif sb.type == SpeechPart.TYPE_SPEECH_P:
                result += ' '
            elif sb.type == SpeechPart.TYPE_SPEECH_LINE:
                result += '\n'
            elif sb.type == SpeechPart.TYPE_STAGE_DIRECTION or sb.type == SpeechPart.TYPE_STAGE_DIRECTION_COUNT:
                result += '\n'
            elif sb.type == SpeechPart.TYPE_STAGE_DIRECTION_INLINE:
                result += ' '

            result += f'{sb.text}'

        return result

    def get_speaker(self) -> str:
        if self.speaker:
            return f'\n{self.speaker}'
        else:
            return ''

    def get_tokens(self):
        result = []
        for sb in self.speech_parts:
            result.extend(sb.tokens)
        return result

    def get_markups(self) -> List[Markup]:
        markups = []

        if self.speaker_markup:
            markups.append(self.speaker_markup)

        for speech_part in self.speech_parts:
            markups.extend(speech_part.markups)

        return markups

    def add_speech_part(self, text: str, speech_type: int, markups: List[Markup]) -> None:
        self.speech_parts.append(SpeechPart(text, speech_type, markups))

    def __len__(self):
        return sum([len(x) for x in self.speech_parts])



from dataclasses import dataclass, field
from typing import List, Tuple

from dramatist.core.DramatistParseException import DramatistParseException
from dramatist.drama.Markup import Markup
from dramatist.drama.Scene import Scene
from dramatist.TextBlock import TextBlock


@dataclass
class Act:
    title: str = ''
    epigraph = ''
    scenes: List[Scene] = field(default_factory=list)
    stage_directions: List[str] = field(default_factory=list)
    start: int = 0
    end: int = 0
    markup: Markup = None

    def add_text_to_epigraph(self, text: str):
        if self.epigraph:
            self.epigraph += '\n'
        self.epigraph += text

    def new_scene(self):
        self.scenes.append(Scene())

    def new_speech(self):
        self.scenes[-1].new_speech()

    def set_scene_title(self, title: str):
        self.scenes[-1].title = title

    def set_speaker(self, speaker: str):
        self.scenes[-1].set_speaker(speaker)

    def add_stage_direction(self, stage_direction: str):
        if len(self.scenes) > 0:
            raise DramatistParseException('Stage direction between scenes')

        self.stage_directions.append(stage_direction)

    def add_scene_stage_direction(self, stage_direction: str):
        self.scenes[-1].add_stage_direction(stage_direction)

    def add_speech(self, speech: str, speech_type: int, markups: List[Markup]):
        self.scenes[-1].add_speech(speech, speech_type, markups)

    def get_pre_scene_length(self) -> int:
        length = 0

        if self.title:
            length += len(self.title) + 1

        if self.epigraph:
            length += len(self.epigraph) + 1

        for stage in self.stage_directions:
            length += len(stage) + 1

        return length

    def get_text(self) -> str:
        text, _ = self.get_text_with_markup()
        return text

    def get_text_with_markup(self) -> Tuple[str, List[Markup]]:
        result = ''
        markups = []

        if self.markup:
            markups.append(self.markup)

        # dummy acts do not have a title
        if self.title:
            result = f'{self.title}'

        if self.epigraph:
            result += f'\n{self.epigraph}'

        for stage in self.stage_directions:
            result += f'\n{stage}'

        for scene in self.scenes:
            if result:
                result += '\n'

            sub_text, sub_markups = scene.get_text_with_markup()
            markups.extend(sub_markups)
            result += f'{sub_text}'

        return result, markups

    def get_text_for_scene(self, scene) -> str:
        return self.scenes[scene-1].get_text()

    def get_text_for_scene_in_blocks(self, scene: int, max_length: int) -> List[TextBlock]:
        return self.scenes[scene-1].get_text_in_blocks(max_length)

    def get_text_for_scene_by_speaker(self, scene: int, max_length: int) -> List[TextBlock]:
        return self.scenes[scene-1].get_text_by_speaker(max_length)

    def get_range(self) -> Tuple[int, int]:
        return self.start, self.end

    def __len__(self):
        return len(self.scenes)



from dramatist.drama.Drama import Drama
import re
import requests
import asyncio

from dramatist.core.DramatistParseException import DramatistParseException
from dramatist.drama.Markup import Markup
from dramatist.drama.SpeechPart import SpeechPart
import xml.etree.ElementTree as ElementTree


class Dramatist:

    def from_file(self, file_path):
        """
        Load a drama from the given file path.
        :param file_path: The file path to the xml file
        :return: A drama
        """
        xml_root = ElementTree.parse(file_path).getroot()
        return self.__create_drama(xml_root)

    def from_api_by_name(self, corpus_name, play_name):
        """
        Load a drama from the DraCor API.
        :param corpus_name: The DraCor corpus name
        :param play_name: The DraCor play name
        :return: A drama
        """
        task1 = asyncio.create_task(
            self.__load_play_from_api(corpus_name, play_name)
        )

        return task1

    def from_api_by_id(self, play_id):
        """
        Load a drama from the DraCor API.
        :param play_id: The DraCor play id
        :return: A drama
        """
        task1 = asyncio.create_task(
            self.__load_play_from_id(play_id)
        )

        return task1

    async def __load_play_from_id(self, play_id):
        play_info = await self.__load_info(play_id)

        if play_info:
            corpus_name = play_info['corpus']
            play_name = play_info['name']
            return await self.__load_play_from_api(corpus_name, play_name)
        else:
            return None

    async def __load_info(self, play_id):
        headers = {
            'Accept': 'application/json'
        }

        response = requests.get(f'https://dracor.org/api/v1/id/{play_id}', headers=headers)
        if response.status_code == 200:
            result_json = response.json()
            return result_json
        else:
            return None

    async def __load_play_from_api(self, corpus_name, play_name):
        headers = {
            'Accept': 'application/xml'
        }

        response = requests.get(f'https://dracor.org/api/v1/corpora/{corpus_name}/plays/{play_name}/tei',
                                headers=headers)

        if response.status_code == 200:
            xml_root = ElementTree.fromstring(response.text)
            return self.__create_drama(xml_root)
        else:
            return None

    def __create_drama(self, xml_root):
        ns = self.__get_namespace(xml_root)
        body_content = xml_root.find(f'.//{ns}body')
        drama_name = xml_root.find(f'.//{ns}title[@type="main"]').text
        drama = Drama(drama_name)
        self.__get_text_from_element(ns, body_content, False, True, [], None, drama,
                                     False)
        drama.update_positions()
        return drama

    def __get_namespace(self, element):
        m = re.match(r'\{.*}', element.tag)
        return m.group(0) if m else ''

    def __get_text_from_element(self, ns, element, use_tail, is_first_act, stack, sibling_tag, drama, first_child):
        new_beginning = False
        result = ''
        markups = []

        if element.tag == f'{ns}div':
            div_type = element.attrib.get('type')
            stack.append(div_type)
        else:
            stack.append(element.tag.removeprefix(ns))

        if element.tag == f'{ns}div':
            div_type = element.attrib.get('type')

            if div_type == 'prologue':
                drama.new_act()

            if div_type == 'act':
                drama.new_act()

            if div_type == 'scene':
                drama.new_scene()

        if element.tag == f'{ns}sp':
            drama.new_speech()

        # There can be speech with only stage instruction before
        if element.tag == f'{ns}lg' and 'sp' not in stack:
            drama.new_speech()

        if element.tag == f'{ns}l' or element.tag == f'{ns}head' or element.tag == f'{ns}stage':
            new_beginning = True

        # emph is ignored, so we need a space
        if element.tag == f'{ns}emph':
            if not first_child:
                result += ' '
            markups.append(Markup(0, -1, 'emph'))

        if element.tag == f'{ns}pb':
            if not first_child:
                result += ' '

        if element.text:
            text = element.text
            if text:
                cleaned_text = self.__clean_text(text)
                if cleaned_text:
                    new_beginning = False
                    result += cleaned_text

        sub_sibling_tag = None
        is_first = True
        for e_pos, text in enumerate(element):
            is_first_child_at_start = False

            if is_first:
                if 'p' == stack[-1] and not result:
                    is_first_child_at_start = True
                elif new_beginning:
                    is_first_child_at_start = True

            if 'p' in stack and 'p' == stack[-1]:
                if result:
                    if is_first:
                        drama.add_speech(result.strip(), SpeechPart.TYPE_SPEECH_P_B, markups)
                        markups = []
                    else:
                        drama.add_speech(result.strip(), SpeechPart.TYPE_SPEECH_P, markups)
                        markups = []

                    result = ''

            inner_text, is_first_act, sub_sibling_tag, new_markups = (
                self.__get_text_from_element(ns, text, True, is_first_act, stack, sub_sibling_tag, drama,
                                             is_first_child_at_start))

            for nm in new_markups:
                if result:
                    nm.start = nm.start + len(result) + 1
                    nm.end = nm.end + len(result) + 1
                else:
                    nm.start = nm.start
                    nm.end = nm.end

            markups.extend(new_markups)
            result += inner_text
            is_first = False

        if 'emph' == stack[-1]:
            length = len(result.strip())
            markups[-1].end = length

        if 'l' in stack and 'l' == stack[-1]:
            if 'epigraph' in stack:
                drama.acts[-1].add_text_to_epigraph(result.strip())
            else:
                drama.add_speech(result.strip(), SpeechPart.TYPE_SPEECH_LINE, markups)
                markups = []

            result = ''

        # there can be quotes
        if 'bibl' == stack[-1]:
            drama.acts[-1].add_text_to_epigraph(result.strip())
            result = ''

        if 'p' in stack and 'p' == stack[-1] and 'sp' in stack:
            if result:
                if sub_sibling_tag:
                    drama.add_speech(result.strip(), SpeechPart.TYPE_SPEECH_P, markups)
                    markups = []
                else:
                    drama.add_speech(result.strip(), SpeechPart.TYPE_SPEECH_P_B, markups)
                    markups = []

            result = ''

        if 'speaker' in stack:
            drama.set_speaker(result)
            result = ''

        if 'stage' in stack and 'stage' == stack[-1]:
            if 'p' in stack:
                if first_child:
                    # TODO: improve special case, see SpeechPart.py
                    drama.add_speech(result.strip(), SpeechPart.TYPE_STAGE_DIRECTION_COUNT, markups)
                    markups = []
                else:
                    drama.add_speech(result.strip(), SpeechPart.TYPE_STAGE_DIRECTION_INLINE, markups)
                    markups = []
                result = ''
            elif 'sp' in stack:
                if (sibling_tag == 'lg' or sibling_tag == 'l' or sibling_tag == 'head' or sibling_tag == 'stage' or
                        sibling_tag == 'p'):
                    drama.add_speech(result.strip(), SpeechPart.TYPE_STAGE_DIRECTION, markups)
                    markups = []
                    result = ''
                else:
                    drama.add_speech(result.strip(), SpeechPart.TYPE_STAGE_DIRECTION_INLINE, markups)
                    markups = []
                    result = ''
            elif 'scene' in stack:
                drama.add_scene_stage_direction(result)
            elif 'act' in stack:
                drama.add_act_stage_direction(result)
            elif 'prologue' in stack:
                drama.add_act_stage_direction(result)
            else:
                raise DramatistParseException('Unknown stage case')

        if 'scene' in stack and 'head' in stack and 'head' == stack[-1]:
            drama.set_scene_title(result)
            result = ''
        elif 'prologue' in stack and 'head' in stack and 'head' == stack[-1]:
            drama.set_act_title(result)
            result = ''
        elif 'act' in stack and 'head' in stack and 'head' == stack[-1]:
            drama.set_act_title(result)
            result = ''

        if 'trailer' in stack:
            drama.set_trailer(result)
            result = ''

        if use_tail and element.tail:
            text = element.tail
            if text:
                if 'p' in stack and 'stage' == stack[-1]:
                    result += ' '

                if 'emph' in stack:
                    result += ' '

                result += self.__clean_text(text)
                result = result.rstrip(' ')

        last_tag = stack.pop()

        return result, is_first_act, last_tag, markups

    def __clean_text(self, text):
        result = re.sub(" +", " ", text, flags=re.DOTALL)

        if not result == ' ':
            result = text.strip()

        result = re.sub(" *\n *", " ", result, flags=re.DOTALL)
        # result = re.sub(" +", " ", result, flags=re.DOTALL)
        return result



from dataclasses import dataclass, field
from typing import List, Tuple

from dramatist.TextBlock import TextBlock
from dramatist.drama.Act import Act
from dramatist.core.DramatistParseException import DramatistParseException
from dramatist.drama.Markup import Markup
from dramatist.drama.Speech import Speech
from dramatist.drama.SpeechPart import SpeechPart
from dramatist.drama.StageDirection import StageDirection


@dataclass
class Drama:
    title: str
    acts: List[Act] = field(default_factory=list)
    trailer: str = ''

    def new_act(self):
        self.acts.append(Act())

    def new_scene(self):
        if len(self.acts) == 0:
            print('Creating dummy act')
            self.new_act()

        self.acts[-1].new_scene()

    def new_speech(self):
        self.acts[-1].new_speech()

    def set_act_title(self, title: str):
        self.acts[-1].title = title

    def set_scene_title(self, title: str):
        self.acts[-1].set_scene_title(title)

    def set_speaker(self, speaker: str):
        self.acts[-1].set_speaker(speaker)

    def add_act_stage_direction(self, stage_direction: str):
        self.acts[-1].add_stage_direction(stage_direction)

    def add_scene_stage_direction(self, stage_direction: str):
        self.acts[-1].add_scene_stage_direction(stage_direction)

    def add_speech(self, speech: str, speech_type: int, markups: List[Markup]):
        self.acts[-1].add_speech(speech, speech_type, markups)

    def set_trailer(self, trailer: str):
        self.trailer = trailer

    @property
    def act_count(self) -> int:
        return len(self.acts)

    @property
    def scene_count(self) -> int:
        return sum([len(x) for x in self.acts])

    def get_text(self) -> str:
        """
        Get the text of the drama.
        :return: The text of the drama.
        """
        text, _ = self.get_text_with_markup()
        return text

    def get_text_with_markup(self) -> Tuple[str, List[Markup]]:
        """
        Get the text of the drama a list of markups.
        :return: A tuple consisting of the drama text and a list of markup objects, i.e. positions of emphasized text,
        speaker names, act and scene titles.
        """
        result = ''
        markups = []

        for act in self.acts:
            if result:
                result += '\n'

            sub_result, sub_markups = act.get_text_with_markup()
            markups.extend(sub_markups)

            result += f'{sub_result}'

        if self.trailer:
            result += f'\n{self.trailer}'

        return result, markups

    def get_structure(self) -> str:
        """
        Get the structure of the drama.
        :return:
        The structure of the drama as a string, for example:
        1. Act (0:500)
        1. Scene (10:50)
        where the numbers in parentheses represent the character start and end positions.
        """
        result = f'Overview: {self.act_count} acts with {self.scene_count} scenes.'

        for act_nr, act in enumerate(self.acts):
            result += f'\n{act_nr+1}. Act ({act.start}:{act.end})'

            for scene_nr, scene in enumerate(act.scenes):
                result += f'\n{scene_nr+1}. Scene ({scene.start}:{scene.end})'

        return result

    def get_text_for_scene(self, act, scene) -> str:
        """
        Get the text for the given scene in the given act.
        :param act: Number of the act
        :param scene: Number of the scene
        :return: The text of the scene
        """
        return self.acts[act-1].get_text_for_scene(scene)

    def get_text_for_scene_in_blocks(self, act, scene, max_length) -> List[TextBlock]:
        """
        Get a list of text blocks for the given scene in the given act. Blocks that are longer than the given
        maximum length (in words) are split into multiple blocks.
        :param act: The act number
        :param scene: The scene number
        :param max_length: The maximum length in words of a returned text block. Longer blocks will be split.
        :return: A list of text blocks
        """
        return self.acts[act-1].get_text_for_scene_in_blocks(scene, max_length)

    def get_text_for_scene_by_speaker(self, act, scene, max_length) -> List[TextBlock]:
        """
        Get a list of text blocks by speaker for the given scene in the given act. This means that for every speaker
        change a new text block is created. Blocks that are longer than the given maximum length (in words)
        are split into multiple blocks.
        :param act: The act number
        :param scene: The scene number
        :param max_length: The maximum length of a returned text block. Longer blocks will be split.
        :return: A list of text blocks
        """
        return self.acts[act-1].get_text_for_scene_by_speaker(scene, max_length)

    def get_scene_act_for_position(self, search_start: int) -> Tuple[int, int]:
        """
        Return a tuple of act and scene number for the given position.
        :param search_start: The character position for which to return the act and scene number
        :return: A tuple of act and scene number. (-1, -1), if act and scene could not be found, or (act_nr, -1) if act
        could be found but scene could not.
        """
        found_act_nr = -1
        for act_pos, act in enumerate(self.acts):
            act_start, act_end = act.get_range()

            if act_start <= search_start < act_end:
                found_act_nr = act_pos + 1
                for scene_pos, scene in enumerate(act.scenes):
                    scene_start, scene_end = scene.get_range()

                    if scene_start <= search_start < scene_end:
                        return act_pos + 1, scene_pos + 1

        return found_act_nr, -1

    def get_scene_act_for_line(self, line_nr: int) -> Tuple[int, int]:
        """
        Return a tuple of act and scene number for the given line number.
        :param line_nr: A line number
        :return: A tuple of act and scene number. (-1, -1), if act and scene could not be found
        """
        for act_pos, act in enumerate(self.acts):
            if act.scenes[-1].end_line >= line_nr:
                for scene_pos, scene in enumerate(act.scenes):
                    if scene.start_line <= line_nr <= scene.end_line:
                        return act_pos + 1, scene_pos + 1

        return -1, -1

    def get_scene_act_for_text(self, text_to_search: str) -> Tuple[int, int]:
        """
        Return a tuple of act and scene number in which the given text appears.
        :param text_to_search: The text so search for
        :return: A tuple of act and scene number. (-1, -1), if act and scene could not be found
        """
        search_string = text_to_search[0:100]
        search_string = (search_string.lower().replace('\n', '').replace('\t', '')
                         .replace(' ', ''))

        drama_text = (self.get_text().lower().replace('\n', '').replace('\t', '')
                      .replace(' ', ''))
        match_cnt = drama_text.count(search_string)

        if match_cnt == 0:
            return -1, -1
        elif match_cnt > 1:
            raise Exception('Act and scene could not be clearly identified!')

        for act_pos, act in enumerate(self.acts):
            for scene_pos, scene in enumerate(act.scenes):
                scene_text = scene.get_text()
                scene_text = (scene_text.lower().replace('\n', '').replace('\t', '')
                              .replace(' ', ''))

                if search_string in scene_text:
                    return act_pos + 1, scene_pos + 1

        return -1, -1

    def get_range_for_scene(self, act: int, scene: int) -> Tuple[int, int]:
        """
        Return a tuple of character start and end position for the given act and scene number.
        :param act: The act number
        :param scene: The scene number
        :return: A tuple of character start and end position
        """
        if act < 1:
            raise ValueError('Act must be greater than 0.')

        if scene < 1:
            raise ValueError('Scene must be greater than 0.')

        return self.acts[act-1].scenes[scene-1].get_range()

    def update_positions(self):
        total_lines = 0
        total_length = 0

        for act in self.acts:
            if act.title:
                act.markup = Markup(total_length, total_length + len(act.title), 'act')
            act.start = total_length
            total_length += act.get_pre_scene_length()
            for scene in act.scenes:
                text = scene.get_text()
                scene.start = total_length
                scene.end = scene.start + len(text)
                scene.start_line = total_lines + 1
                if scene.title:
                    scene.markup = Markup(total_length, total_length + len(scene.title), 'scene')
                    total_length += len(scene.title) + 1

                for scene_part in scene.scene_parts:
                    if isinstance(scene_part, StageDirection):
                        total_length += len(scene_part.text) + 1
                        continue

                    if not isinstance(scene_part, Speech):
                        raise DramatistParseException('Unknown scene part type')

                    scene_part.start_line = total_lines + 1
                    if scene_part.speaker:
                        scene_part.speaker_markup = Markup(total_length, total_length + len(scene_part.speaker),
                                                           'speaker')
                        total_length += len(scene_part.speaker) + 1
                    for speech_part in scene_part.speech_parts:
                        line_count = 0

                        for markup in speech_part.markups:
                            markup.start = total_length + markup.start
                            markup.end = total_length + markup.end

                        if speech_part.type == SpeechPart.TYPE_SPEECH_LINE:
                            line_count += 1
                        elif speech_part.type == SpeechPart.TYPE_SPEECH_P_B:
                            line_count += 1
                        elif (speech_part.type == SpeechPart.TYPE_STAGE_DIRECTION or
                              speech_part.type == SpeechPart.TYPE_STAGE_DIRECTION_COUNT):
                            speech_part.markups.append(Markup(total_length, total_length +
                                                              len(speech_part.text), 'stage'))

                            if speech_part.type == SpeechPart.TYPE_STAGE_DIRECTION_COUNT:
                                line_count += 1

                        elif speech_part.type == SpeechPart.TYPE_STAGE_DIRECTION_INLINE:
                            speech_part.markups.append(Markup(total_length, total_length +
                                                              len(speech_part.text), 'stage_inline'))

                        scene_part.end_line = total_lines + line_count
                        total_lines += line_count
                        total_length += len(speech_part.text) + 1

                scene.end_line = total_lines
            act.end = total_length - 1

    def __len__(self) -> int:
        return len(self.acts)

    def __str__(self) -> str:
        return self.get_text()



from argparse import ArgumentParser
from os.path import join

from dramatist.drama.Act import Act
from dramatist.core.Dramatist import Dramatist
import asyncio
import json

from dramatist.core.DramatistParseException import DramatistParseException
from dramatist.drama.Scene import Scene

from requests.exceptions import RequestException


def __json_encoder_drama(obj):
    if isinstance(obj, (Act, Scene)):
        result_dict = obj.__dict__
        return result_dict

    return obj.__dict__


async def main():
    argument_parser = ArgumentParser()

    argument_parser.add_argument('--drama-path', dest='drama_path',
                                 help='Path to the drama xml file', required=False)
    argument_parser.add_argument('--corpus-name', dest='corpus_name',
                                 help='Name of the corpus', required=False)
    argument_parser.add_argument('--play-name', dest='play_name', help='Name of the play', required=False)
    argument_parser.add_argument('--play-id', dest='play_id', help='ID of the play', required=False)
    argument_parser.add_argument('--output-type', choices=['json', 'plain', 'structure'],
                                 dest="output_type", default="text", help="The output type")
    argument_parser.add_argument('--output-folder-path', dest="output_folder_path",
                                 help="The output folder path. If this option is set the output will be saved to a file"
                                      " created in the specified folder")

    args = argument_parser.parse_args()
    drama_path = args.drama_path
    corpus_name = args.corpus_name
    play_name = args.play_name
    play_id = args.play_id
    output_type = args.output_type
    output_folder_path = args.output_folder_path

    if not drama_path and not corpus_name and not play_name and not play_id:
        argument_parser.error('--drama-path or --corpus-name and --play-name or --play-id must be set')
    if drama_path and (corpus_name or play_name or play_id):
        argument_parser.error('only --drama-path or (--corpus-name and --play-name) or --play-id can be used at a time')
    elif play_id and (drama_path or corpus_name or play_name):
        argument_parser.error('only --drama-path or (--corpus-name and --play-name) or --play-id can be used at a time')
    elif (corpus_name and not play_name) or (play_name and not corpus_name):
        argument_parser.error('--corpus-name and --play-name have both to be set together')

    dramatist = Dramatist()

    try:
        if drama_path:
            drama = dramatist.from_file(drama_path)
        elif play_id:
            drama = await dramatist.from_api_by_id(play_id)
        else:
            drama = await dramatist.from_api_by_name(corpus_name, play_name)
    except RequestException as e:
        print(f'Could not load drama: {e}')
    except DramatistParseException as e:
        print(f'Could not parse drama: {e}')
    else:
        if drama:
            output_string = ''
            file_type = ''

            if output_type == 'text':
                output_string = drama.get_text()
                file_type = 'txt'
            elif output_type == 'json':
                output_string = json.dumps(drama, default=__json_encoder_drama)
                file_type = 'json'
            elif output_type == 'structure':
                output_string = drama.get_structure()
                file_type = 'txt'

            if output_folder_path:
                with open(join(output_folder_path, f'{drama.title}.{file_type}'), 'w', encoding='utf-8') as output_file:
                    output_file.write(output_string)
            else:
                print(output_string)
        else:
            print('Could not load drama')


def launch():
    asyncio.run(main())


if __name__ == '__main__':
    launch()



from typing import List
from dramatist.drama.Markup import Markup


class SpeechPart:
    TYPE_SPEECH_LINE = 0
    TYPE_SPEECH_P_B = 1
    TYPE_SPEECH_P = 2
    TYPE_STAGE_DIRECTION = 3
    TYPE_STAGE_DIRECTION_INLINE = 4
    # For counting, we run into issues, when a p-block starts with a stage direction. So we need a special type
    # TODO: Could this be improved?
    TYPE_STAGE_DIRECTION_COUNT = 5

    _text: str = ''

    def __init__(self, text: str, speech_type: int, markups: List[Markup]):
        self.type = speech_type
        self.tokens: List[str] = list()
        self.length = 0
        self.text = text
        self.markups = markups

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, v: str):
        self._text = v
        self.tokens = v.split(' ')
        self.length = len(self.tokens)

    def __len__(self):
        return self.length
